# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This implementation has been borrowed from kserve/kserve repository
# https://github.com/kserve/kserve/blob/release-0.11/python/kserve/kserve/protocol/infer_type.py

import struct
from typing import Dict, List, Optional

import numpy
import numpy as np
import pandas as pd
from hopsworks_common.client.istio.grpc.errors import InvalidInput
from hopsworks_common.client.istio.grpc.proto.grpc_predict_v2_pb2 import (
    InferTensorContents,
    ModelInferRequest,
    ModelInferResponse,
)
from hopsworks_common.client.istio.utils.numpy_codec import from_np_dtype, to_np_dtype


GRPC_CONTENT_DATATYPE_MAPPINGS = {
    "BOOL": "bool_contents",
    "INT8": "int_contents",
    "INT16": "int_contents",
    "INT32": "int_contents",
    "INT64": "int64_contents",
    "UINT8": "uint_contents",
    "UINT16": "uint_contents",
    "UINT32": "uint_contents",
    "UINT64": "uint64_contents",
    "FP32": "fp32_contents",
    "FP64": "fp64_contents",
    "BYTES": "bytes_contents",
}


def raise_error(msg):
    """
    Raise error with the provided message
    """
    raise InferenceServerException(msg=msg) from None


def serialize_byte_tensor(input_tensor):
    """
    Serializes a bytes tensor into a flat numpy array of length prepended
    bytes. The numpy array should use dtype of np.object. For np.bytes,
    numpy will remove trailing zeros at the end of byte sequence and because
    of this it should be avoided.

    Parameters
    ----------
    input_tensor : np.array
        The bytes tensor to serialize.

    Returns
    -------
    serialized_bytes_tensor : np.array
        The 1-D numpy array of type uint8 containing the serialized bytes in row-major form.

    Raises
    ------
    InferenceServerException
        If unable to serialize the given tensor.
    """

    if input_tensor.size == 0:
        return np.empty([0], dtype=np.object_)

    # If the input is a tensor of string/bytes objects, then must flatten those into
    # a 1-dimensional array containing the 4-byte byte size followed by the
    # actual element bytes. All elements are concatenated together in row-major
    # order.

    if (input_tensor.dtype != np.object_) and (input_tensor.dtype.type != np.bytes_):
        raise_error("cannot serialize bytes tensor: invalid datatype")

    flattened_ls = []
    # 'C' order is row-major.
    for obj in np.nditer(input_tensor, flags=["refs_ok"], order="C"):
        # If directly passing bytes to BYTES type,
        # don't convert it to str as Python will encode the
        # bytes which may distort the meaning
        if input_tensor.dtype == np.object_:
            if isinstance(obj.item(), bytes):
                s = obj.item()
            else:
                s = str(obj.item()).encode("utf-8")
        else:
            s = obj.item()
        flattened_ls.append(struct.pack("<I", len(s)))
        flattened_ls.append(s)
    flattened = b"".join(flattened_ls)
    flattened_array = np.asarray(flattened, dtype=np.object_)
    if not flattened_array.flags["C_CONTIGUOUS"]:
        flattened_array = np.ascontiguousarray(flattened_array, dtype=np.object_)
    return flattened_array


class InferenceServerException(Exception):
    """Exception indicating non-Success status.

    Parameters
    ----------
    msg : str
        A brief description of error

    status : str
        The error code

    debug_details : str
        The additional details on the error

    """

    def __init__(self, msg, status=None, debug_details=None):
        self._msg = msg
        self._status = status
        self._debug_details = debug_details

    def __str__(self):
        msg = super().__str__() if self._msg is None else self._msg
        if self._status is not None:
            msg = "[" + self._status + "] " + msg
        return msg

    def message(self):
        """Get the exception message.

        Returns
        -------
        str
            The message associated with this exception, or None if no message.

        """
        return self._msg

    def status(self):
        """Get the status of the exception.

        Returns
        -------
        str
            Returns the status of the exception

        """
        return self._status

    def debug_details(self):
        """Get the detailed information about the exception
        for debugging purposes

        Returns
        -------
        str
            Returns the exception details

        """
        return self._debug_details


class InferInput:
    _name: str
    _shape: List[int]
    _datatype: str
    _parameters: Dict

    def __init__(self, name, shape, datatype, data=None, parameters=None):
        """An object of InferInput class is used to describe
        input tensor for an inference request.
        Parameters
        ----------
        name : str
            The name of input whose data will be described by this object
        shape : list
            The shape of the associated input.
        datatype : str
            The datatype of the associated input.
        data : Union[List, InferTensorContents]
            The data of the REST/gRPC input. When data is not set, raw_data is used for gRPC for numpy array bytes.
        parameters : dict
            The additional server-specific parameters.
        """
        if parameters is None:
            parameters = {}
        self._name = name
        self._shape = shape
        self._datatype = datatype
        self._parameters = parameters
        self._data = data
        self._raw_data = None

    @property
    def name(self):
        """Get the name of input associated with this object.
        Returns
        -------
        str
            The name of input
        """
        return self._name

    @property
    def datatype(self):
        """Get the datatype of input associated with this object.
        Returns
        -------
        str
            The datatype of input
        """
        return self._datatype

    @property
    def data(self):
        """Get the data of InferInput"""
        return self._data

    @property
    def shape(self):
        """Get the shape of input associated with this object.
        Returns
        -------
        list
            The shape of input
        """
        return self._shape

    @property
    def parameters(self):
        """Get the parameters of input associated with this object.
        Returns
        -------
        dict
            The key, value pair of string and InferParameter
        """
        return self._parameters

    def set_shape(self, shape):
        """Set the shape of input.
        Parameters
        ----------
        shape : list
            The shape of the associated input.
        """
        self._shape = shape

    def as_numpy(self) -> np.ndarray:
        dtype = to_np_dtype(self.datatype)
        if dtype is None:
            raise InvalidInput("invalid datatype in the input")
        if self._raw_data is not None:
            np_array = np.frombuffer(self._raw_data, dtype=dtype)
            return np_array.reshape(self._shape)
        else:
            np_array = np.array(self._data, dtype=dtype)
            return np_array.reshape(self._shape)

    def set_data_from_numpy(self, input_tensor, binary_data=True):
        """Set the tensor data from the specified numpy array for
        input associated with this object.
        Parameters
        ----------
        input_tensor : numpy array
            The tensor data in numpy array format
        binary_data : bool
            Indicates whether to set data for the input in binary format
            or explicit tensor within JSON. The default value is True,
            which means the data will be delivered as binary data in the
            HTTP body after the JSON object.
        Raises
        ------
        InferenceServerException
            If failed to set data for the tensor.
        """
        if not isinstance(input_tensor, (np.ndarray,)):
            raise_error("input_tensor must be a numpy array")

        dtype = from_np_dtype(input_tensor.dtype)
        if self._datatype != dtype:
            raise_error(
                "got unexpected datatype {} from numpy array, expected {}".format(
                    dtype, self._datatype
                )
            )
        valid_shape = True
        if len(self._shape) != len(input_tensor.shape):
            valid_shape = False
        else:
            for i in range(len(self._shape)):
                if self._shape[i] != input_tensor.shape[i]:
                    valid_shape = False
        if not valid_shape:
            raise_error(
                "got unexpected numpy array shape [{}], expected [{}]".format(
                    str(input_tensor.shape)[1:-1], str(self._shape)[1:-1]
                )
            )

        if not binary_data:
            self._parameters.pop("binary_data_size", None)
            self._raw_data = None
            if self._datatype == "BYTES":
                self._data = []
                try:
                    if input_tensor.size > 0:
                        for obj in np.nditer(
                            input_tensor, flags=["refs_ok"], order="C"
                        ):
                            # We need to convert the object to string using utf-8,
                            # if we want to use the binary_data=False. JSON requires
                            # the input to be a UTF-8 string.
                            if input_tensor.dtype == np.object_:
                                if isinstance(obj.item(), bytes):
                                    self._data.append(str(obj.item(), encoding="utf-8"))
                                else:
                                    self._data.append(str(obj.item()))
                            else:
                                self._data.append(str(obj.item(), encoding="utf-8"))
                except UnicodeDecodeError:
                    raise_error(
                        f'Failed to encode "{obj.item()}" using UTF-8. Please use binary_data=True, if'
                        " you want to pass a byte array."
                    )
            else:
                self._data = [val.item() for val in input_tensor.flatten()]
        else:
            self._data = None
            if self._datatype == "BYTES":
                serialized_output = serialize_byte_tensor(input_tensor)
                if serialized_output.size > 0:
                    self._raw_data = serialized_output.item()
                else:
                    self._raw_data = b""
            else:
                self._raw_data = input_tensor.tobytes()
            self._parameters["binary_data_size"] = len(self._raw_data)


def get_content(datatype: str, data: InferTensorContents):
    if datatype == "BOOL":
        return list(data.bool_contents)
    elif datatype in ["UINT8", "UINT16", "UINT32"]:
        return list(data.uint_contents)
    elif datatype == "UINT64":
        return list(data.uint64_contents)
    elif datatype in ["INT8", "INT16", "INT32"]:
        return list(data.int_contents)
    elif datatype == "INT64":
        return list(data.int64_contents)
    elif datatype == "FP32":
        return list(data.fp32_contents)
    elif datatype == "FP64":
        return list(data.fp64_contents)
    elif datatype == "BYTES":
        return list(data.bytes_contents)
    else:
        raise InvalidInput("invalid content type")


class InferRequest:
    """InferenceRequest Model

    $inference_request =
    {
      "id" : $string #optional,
      "parameters" : $parameters #optional,
      "inputs" : [ $request_input, ... ],
      "outputs" : [ $request_output, ... ] #optional
    }
    """

    id: Optional[str]
    model_name: str
    parameters: Optional[Dict]
    inputs: List[InferInput]
    from_grpc: bool

    def __init__(
        self,
        model_name: str,
        infer_inputs: List[InferInput],
        request_id=None,
        raw_inputs=None,
        from_grpc=False,
        parameters=None,
    ):
        if parameters is None:
            parameters = {}
        self.id = request_id
        self.model_name = model_name
        self.inputs = infer_inputs
        self.parameters = parameters
        self.from_grpc = from_grpc
        if raw_inputs:
            for i, raw_input in enumerate(raw_inputs):
                self.inputs[i]._raw_data = raw_input

    @classmethod
    def from_grpc(cls, request: ModelInferRequest):
        infer_inputs = [
            InferInput(
                name=input_tensor.name,
                shape=list(input_tensor.shape),
                datatype=input_tensor.datatype,
                data=get_content(input_tensor.datatype, input_tensor.contents),
                parameters=input_tensor.parameters,
            )
            for input_tensor in request.inputs
        ]
        return cls(
            request_id=request.id,
            model_name=request.model_name,
            infer_inputs=infer_inputs,
            raw_inputs=request.raw_input_contents,
            from_grpc=True,
            parameters=request.parameters,
        )

    def to_rest(self) -> Dict:
        """Converts the InferRequest object to v2 REST InferenceRequest message"""
        infer_inputs = []
        for infer_input in self.inputs:
            infer_input_dict = {
                "name": infer_input.name,
                "shape": infer_input.shape,
                "datatype": infer_input.datatype,
            }
            if isinstance(infer_input.data, numpy.ndarray):
                infer_input.set_data_from_numpy(infer_input.data, binary_data=False)
                infer_input_dict["data"] = infer_input.data
            else:
                infer_input_dict["data"] = infer_input.data
            infer_inputs.append(infer_input_dict)
        return {"id": self.id, "inputs": infer_inputs}

    def to_grpc(self) -> ModelInferRequest:
        """Converts the InferRequest object to gRPC ModelInferRequest message"""
        infer_inputs = []
        raw_input_contents = []
        for infer_input in self.inputs:
            if isinstance(infer_input.data, numpy.ndarray):
                infer_input.set_data_from_numpy(infer_input.data, binary_data=True)
            infer_input_dict = {
                "name": infer_input.name,
                "shape": infer_input.shape,
                "datatype": infer_input.datatype,
            }
            if infer_input._raw_data is not None:
                raw_input_contents.append(infer_input._raw_data)
            else:
                if not isinstance(infer_input.data, List):
                    raise InvalidInput("input data is not a List")
                infer_input_dict["contents"] = {}
                data_key = GRPC_CONTENT_DATATYPE_MAPPINGS.get(
                    infer_input.datatype, None
                )
                if data_key is not None:
                    infer_input._data = [
                        bytes(val, "utf-8") if isinstance(val, str) else val
                        for val in infer_input.data
                    ]  # str to byte conversion for grpc proto
                    infer_input_dict["contents"][data_key] = infer_input.data
                else:
                    raise InvalidInput("invalid input datatype")
            infer_inputs.append(infer_input_dict)

        return ModelInferRequest(
            id=self.id,
            model_name=self.model_name,
            inputs=infer_inputs,
            raw_input_contents=raw_input_contents,
        )

    def as_dataframe(self) -> pd.DataFrame:
        """
        Decode the tensor inputs as pandas dataframe
        """
        dfs = []
        for input in self.inputs:
            input_data = input.data
            if input.datatype == "BYTES":
                input_data = [
                    str(val, "utf-8") if isinstance(val, bytes) else val
                    for val in input.data
                ]
            dfs.append(pd.DataFrame(input_data, columns=[input.name]))
        return pd.concat(dfs, axis=1)


class InferOutput:
    def __init__(self, name, shape, datatype, data=None, parameters=None):
        """An object of InferOutput class is used to describe
        input tensor for an inference request.
        Parameters
        ----------
        name : str
            The name of input whose data will be described by this object
        shape : list
            The shape of the associated input.
        datatype : str
            The datatype of the associated input.
        data : Union[List, InferTensorContents]
            The data of the REST/gRPC input. When data is not set, raw_data is used for gRPC for numpy array bytes.
        parameters : dict
            The additional server-specific parameters.
        """
        if parameters is None:
            parameters = {}
        self._name = name
        self._shape = shape
        self._datatype = datatype
        self._parameters = parameters
        self._data = data
        self._raw_data = None

    @property
    def name(self):
        """Get the name of input associated with this object.
        Returns
        -------
        str
            The name of input
        """
        return self._name

    @property
    def datatype(self):
        """Get the datatype of input associated with this object.
        Returns
        -------
        str
            The datatype of input
        """
        return self._datatype

    @property
    def data(self):
        """Get the data of InferOutput"""
        return self._data

    @property
    def shape(self):
        """Get the shape of input associated with this object.
        Returns
        -------
        list
            The shape of input
        """
        return self._shape

    @property
    def parameters(self):
        """Get the parameters of input associated with this object.
        Returns
        -------
        dict
            The key, value pair of string and InferParameter
        """
        return self._parameters

    def set_shape(self, shape):
        """Set the shape of input.
        Parameters
        ----------
        shape : list
            The shape of the associated input.
        """
        self._shape = shape

    def as_numpy(self) -> numpy.ndarray:
        """
        Decode the tensor data as numpy array
        """
        dtype = to_np_dtype(self.datatype)
        if dtype is None:
            raise InvalidInput("invalid datatype in the input")
        if self._raw_data is not None:
            np_array = np.frombuffer(self._raw_data, dtype=dtype)
            return np_array.reshape(self._shape)
        else:
            np_array = np.array(self._data, dtype=dtype)
            return np_array.reshape(self._shape)

    def set_data_from_numpy(self, input_tensor, binary_data=True):
        """Set the tensor data from the specified numpy array for
        input associated with this object.
        Parameters
        ----------
        input_tensor : numpy array
            The tensor data in numpy array format
        binary_data : bool
            Indicates whether to set data for the input in binary format
            or explicit tensor within JSON. The default value is True,
            which means the data will be delivered as binary data in the
            HTTP body after the JSON object.
        Raises
        ------
        InferenceServerException
            If failed to set data for the tensor.
        """
        if not isinstance(input_tensor, (np.ndarray,)):
            raise_error("input_tensor must be a numpy array")

        dtype = from_np_dtype(input_tensor.dtype)
        if self._datatype != dtype:
            raise_error(
                "got unexpected datatype {} from numpy array, expected {}".format(
                    dtype, self._datatype
                )
            )
        valid_shape = True
        if len(self._shape) != len(input_tensor.shape):
            valid_shape = False
        else:
            for i in range(len(self._shape)):
                if self._shape[i] != input_tensor.shape[i]:
                    valid_shape = False
        if not valid_shape:
            raise_error(
                "got unexpected numpy array shape [{}], expected [{}]".format(
                    str(input_tensor.shape)[1:-1], str(self._shape)[1:-1]
                )
            )

        if not binary_data:
            self._parameters.pop("binary_data_size", None)
            self._raw_data = None
            if self._datatype == "BYTES":
                self._data = []
                try:
                    if input_tensor.size > 0:
                        for obj in np.nditer(
                            input_tensor, flags=["refs_ok"], order="C"
                        ):
                            # We need to convert the object to string using utf-8,
                            # if we want to use the binary_data=False. JSON requires
                            # the input to be a UTF-8 string.
                            if input_tensor.dtype == np.object_:
                                if isinstance(obj.item(), bytes):
                                    self._data.append(str(obj.item(), encoding="utf-8"))
                                else:
                                    self._data.append(str(obj.item()))
                            else:
                                self._data.append(str(obj.item(), encoding="utf-8"))
                except UnicodeDecodeError:
                    raise_error(
                        f'Failed to encode "{obj.item()}" using UTF-8. Please use binary_data=True, if'
                        " you want to pass a byte array."
                    )
            else:
                self._data = [val.item() for val in input_tensor.flatten()]
        else:
            self._data = None
            if self._datatype == "BYTES":
                serialized_output = serialize_byte_tensor(input_tensor)
                if serialized_output.size > 0:
                    self._raw_data = serialized_output.item()
                else:
                    self._raw_data = b""
            else:
                self._raw_data = input_tensor.tobytes()
            self._parameters["binary_data_size"] = len(self._raw_data)


class InferResponse:
    """InferenceResponse

    $inference_response =
    {
      "model_name" : $string,
      "model_version" : $string #optional,
      "id" : $string,
      "parameters" : $parameters #optional,
      "outputs" : [ $response_output, ... ]
    }
    """

    id: str
    model_name: str
    parameters: Optional[Dict]
    outputs: List[InferOutput]
    from_grpc: bool

    def __init__(
        self,
        response_id: str,
        model_name: str,
        infer_outputs: List[InferOutput],
        raw_outputs=None,
        from_grpc=False,
        parameters=None,
    ):
        if parameters is None:
            parameters = {}
        self.id = response_id
        self.model_name = model_name
        self.outputs = infer_outputs
        self.parameters = parameters
        self.from_grpc = from_grpc
        if raw_outputs:
            for i, raw_output in enumerate(raw_outputs):
                self.outputs[i]._raw_data = raw_output

    @classmethod
    def from_grpc(cls, response: ModelInferResponse) -> "InferResponse":
        infer_outputs = [
            InferOutput(
                name=output.name,
                shape=list(output.shape),
                datatype=output.datatype,
                data=get_content(output.datatype, output.contents),
                parameters=output.parameters,
            )
            for output in response.outputs
        ]
        return cls(
            model_name=response.model_name,
            response_id=response.id,
            parameters=response.parameters,
            infer_outputs=infer_outputs,
            raw_outputs=response.raw_output_contents,
            from_grpc=True,
        )

    @classmethod
    def from_rest(cls, model_name: str, response: Dict) -> "InferResponse":
        infer_outputs = [
            InferOutput(
                name=output["name"],
                shape=list(output["shape"]),
                datatype=output["datatype"],
                data=output["data"],
                parameters=output.get("parameters", {}),
            )
            for output in response["outputs"]
        ]
        return cls(
            model_name=model_name,
            response_id=response.get("id", None),
            parameters=response.get("parameters", {}),
            infer_outputs=infer_outputs,
        )

    def to_rest(self) -> Dict:
        """Converts the InferResponse object to v2 REST InferenceRequest message"""
        infer_outputs = []
        for infer_output in self.outputs:
            infer_output_dict = {
                "name": infer_output.name,
                "shape": infer_output.shape,
                "datatype": infer_output.datatype,
            }
            if isinstance(infer_output.data, numpy.ndarray):
                infer_output.set_data_from_numpy(infer_output.data, binary_data=False)
                infer_output_dict["data"] = infer_output.data
            elif isinstance(infer_output._raw_data, bytes):
                infer_output_dict["data"] = infer_output.as_numpy().tolist()
            else:
                infer_output_dict["data"] = infer_output.data
            infer_outputs.append(infer_output_dict)
        res = {"id": self.id, "model_name": self.model_name, "outputs": infer_outputs}
        return res

    def to_grpc(self) -> ModelInferResponse:
        """Converts the InferResponse object to gRPC ModelInferRequest message"""
        infer_outputs = []
        raw_output_contents = []
        for infer_output in self.outputs:
            if isinstance(infer_output.data, numpy.ndarray):
                infer_output.set_data_from_numpy(infer_output.data, binary_data=True)
            infer_output_dict = {
                "name": infer_output.name,
                "shape": infer_output.shape,
                "datatype": infer_output.datatype,
            }
            if infer_output._raw_data is not None:
                raw_output_contents.append(infer_output._raw_data)
            else:
                if not isinstance(infer_output.data, List):
                    raise InvalidInput("output data is not a List")
                infer_output_dict["contents"] = {}
                data_key = GRPC_CONTENT_DATATYPE_MAPPINGS.get(
                    infer_output.datatype, None
                )
                if data_key is not None:
                    infer_output._data = [
                        bytes(val, "utf-8") if isinstance(val, str) else val
                        for val in infer_output.data
                    ]  # str to byte conversion for grpc proto
                    infer_output_dict["contents"][data_key] = infer_output.data
                else:
                    raise InvalidInput("to_grpc: invalid output datatype")
            infer_outputs.append(infer_output_dict)

        return ModelInferResponse(
            id=self.id,
            model_name=self.model_name,
            outputs=infer_outputs,
            raw_output_contents=raw_output_contents,
        )



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json

import humps
from hopsworks_common import util


class GitCommit:
    def __init__(
        self,
        name=None,
        email=None,
        message=None,
        commit_hash=None,
        time=None,
        type=None,
        href=None,
        expand=None,
        items=None,
        count=None,
        **kwargs,
    ):
        self._name = name
        self._email = email
        self._message = message
        self._hash = commit_hash
        self._time = time
        self._type = type
        self._href = href
        self._expand = expand
        self._items = items
        self._count = count

    @classmethod
    def from_response_json(cls, json_dict):
        if json_dict:
            json_decamelized = humps.decamelize(json_dict)
            if "count" in json_decamelized:
                if json_decamelized["count"] == 0:
                    return []
                return [cls(**commit) for commit in json_decamelized["items"]]
            else:
                return cls(**json_decamelized)
        else:
            return None

    @property
    def name(self):
        """Name of the user"""
        return self._name

    @property
    def email(self):
        """Email of the user"""
        return self._email

    @property
    def message(self):
        """Commit message"""
        return self._message

    @property
    def hash(self):
        """Commit hash"""
        return self._hash

    @property
    def time(self):
        """Timestamp for the commit"""
        return self._time

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def __str__(self):
        return self.json()

    def __repr__(self):
        return f"GitCommit({self._name!r}, {self._message!r}, {self._hash!r})"



#
#   Copyright 2021 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from typing import Optional, Union

import numpy
import pandas
from hopsworks_common import usage
from hsml.model_schema import ModelSchema
from hsml.torch.model import Model


_mr = None


@usage.method_logger
def create_model(
    name: str,
    version: Optional[int] = None,
    metrics: Optional[dict] = None,
    description: Optional[str] = None,
    input_example: Optional[
        Union[pandas.DataFrame, pandas.Series, numpy.ndarray, list]
    ] = None,
    model_schema: Optional[ModelSchema] = None,
    feature_view=None,
    training_dataset_version: Optional[int] = None,
):
    """Create a Torch model metadata object.

    !!! note "Lazy"
        This method is lazy and does not persist any metadata or uploads model artifacts in the
        model registry on its own. To save the model object and the model artifacts, call the `save()` method with a
        local file path to the directory containing the model artifacts.

    # Arguments
        name: Name of the model to create.
        version: Optionally version of the model to create, defaults to `None` and
            will create the model with incremented version from the last
            version in the model registry.
        metrics: Optionally a dictionary with model evaluation metrics (e.g., accuracy, MAE)
        description: Optionally a string describing the model, defaults to empty string
            `""`.
        input_example: Optionally an input example that represents a single input for the model, defaults to `None`.
        model_schema: Optionally a model schema for the model inputs and/or outputs.
        feature_view: Optionally a feature view object returned by querying the feature store. If the feature view is not provided, the model will not have access to provenance.
        training_dataset_version: Optionally a training dataset version. If training dataset version is not provided, but the feature view is provided, the training dataset version used will be the last accessed training dataset of the feature view, within the code/notebook that reads the feature view and training dataset and then creates the model.

    # Returns
        `Model`. The model metadata object.
    """
    model = Model(
        id=None,
        name=name,
        version=version,
        description=description,
        metrics=metrics,
        input_example=input_example,
        model_schema=model_schema,
        feature_view=feature_view,
        training_dataset_version=training_dataset_version,
    )
    model._shared_registry_project_name = _mr.shared_registry_project_name
    model._model_registry_id = _mr.model_registry_id

    return model



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import requests
from hopsworks_common.client import auth
from hopsworks_common.client.istio import base as istio


class Client(istio.Client):
    def __init__(
        self,
        host,
        port,
        project,
        api_key_value,
        hostname_verification=None,
        trust_store_path=None,
        scheme="http",
    ):
        """Initializes a client in an external environment such as AWS Sagemaker."""
        self._host = host
        self._port = port
        self._base_url = f"{scheme}://{self._host}:{str(self._port)}"
        self._project_name = project

        self._auth = auth.ApiKeyAuth(api_key_value)

        self._session = requests.session()
        self._connected = True
        self._verify = self._get_verify(hostname_verification, trust_store_path)

        self._cert_key = None

    def _close(self):
        """Closes a client."""
        self._connected = False

    def replace_public_host(self, url):
        """no need to replace as we are already in external client"""
        return url

    def _is_external(self):
        return True

    @property
    def host(self):
        return self._host



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


from hsfs import feature_group as feature_group_mod
from hsfs import feature_store as feature_store_mod


class TestFeatureStore:
    def test_from_response_json(self, mocker, backend_fixtures):
        # Arrange
        mocker.patch("hopsworks_common.client.get_instance")
        json = backend_fixtures["feature_store"]["get"]["response"]

        # Act
        fs = feature_store_mod.FeatureStore.from_response_json(json)

        # Assert
        assert fs.id == 11
        assert fs.name == "test_featurestore_name"
        assert fs._created == "test_created"
        assert fs.project_name == "test_project_name"
        assert fs.project_id == 67
        assert fs._online_feature_store_name == "test_online_featurestore_name"
        assert fs._online_feature_store_size == 31
        assert fs._offline_feature_store_name == "test_offline_featurestore_name"
        assert fs.online_enabled is True
        assert fs._num_feature_groups == 2
        assert fs._num_training_datasets == 3
        assert fs._num_storage_connectors == 4
        assert fs._num_feature_views == 5

    def test_from_response_json_basic_info(self, mocker, backend_fixtures):
        # Arrange
        mocker.patch("hopsworks_common.client.get_instance")
        json = backend_fixtures["feature_store"]["get_basic_info"]["response"]

        # Act
        fs = feature_store_mod.FeatureStore.from_response_json(json)

        # Assert
        assert fs.id == 11
        assert fs.name == "test_featurestore_name"
        assert fs._created == "test_created"
        assert fs.project_name == "test_project_name"
        assert fs.project_id == 67
        assert fs._online_feature_store_name is None
        assert fs._online_feature_store_size is None
        assert fs._offline_feature_store_name == "test_offline_featurestore_name"
        assert fs.online_enabled is True
        assert fs._num_feature_groups is None
        assert fs._num_training_datasets is None
        assert fs._num_storage_connectors is None
        assert fs._num_feature_views is None

    def test_get_feature_group(self, backend_fixtures, mocker):
        # Arrange
        mocker.patch("hopsworks_common.client.get_instance")
        json = backend_fixtures["feature_store"]["get"]["response"]
        fs = feature_store_mod.FeatureStore.from_response_json(json)
        fg = feature_group_mod.FeatureGroup.from_response_json(
            backend_fixtures["feature_group"]["get"]["response"]
        )
        mocker.patch("hsfs.core.feature_group_api.FeatureGroupApi.get", return_value=fg)

        # Act
        fg_res = fs.get_feature_group("test_feature_group_name")

        # Assert
        assert fg_res.feature_store == fs
        assert fg_res._feature_store == fs

    def test_get_feature_group_by_name_not_found(self, backend_fixtures, mocker):
        # Arrange
        mocker.patch("hopsworks_common.client.get_instance")
        json = backend_fixtures["feature_store"]["get"]["response"]
        fs = feature_store_mod.FeatureStore.from_response_json(json)
        mocker.patch(
            "hsfs.core.feature_group_api.FeatureGroupApi._get_feature_group_by_version",
            return_value=None,
        )

        # Act
        fg_res = fs.get_feature_group("test_feature_group_name")

        # Assert
        assert fg_res is None

    def test_get_feature_groups_not_found(self, backend_fixtures, mocker):
        # Arrange
        mocker.patch("hopsworks_common.client.get_instance")
        json = backend_fixtures["feature_store"]["get"]["response"]
        fs = feature_store_mod.FeatureStore.from_response_json(json)
        mocker.patch(
            "hsfs.core.feature_group_api.FeatureGroupApi._get_feature_group_by_name",
            return_value=[],
        )

        # Act
        fg_res = fs.get_feature_groups("test_feature_group_name")

        # Assert
        assert fg_res == []

    def test_get_feature_group_by_name_and_version_not_found(
        self, backend_fixtures, mocker
    ):
        # Arrange
        mocker.patch("hopsworks_common.client.get_instance")
        json = backend_fixtures["feature_store"]["get"]["response"]
        fs = feature_store_mod.FeatureStore.from_response_json(json)
        mocker.patch(
            "hsfs.core.feature_group_api.FeatureGroupApi._get_feature_group_by_version",
            return_value=None,
        )

        # Act
        fg_res = fs.get_feature_group("test_feature_group_name", version=10)

        # Assert
        assert fg_res is None

    def test_create_feature_group(self, backend_fixtures, mocker):
        # Arrange
        mocker.patch("hopsworks_common.client.get_instance")
        mocker.patch("hsfs.engine.get_type", return_value="python")
        json = backend_fixtures["feature_store"]["get"]["response"]
        fs = feature_store_mod.FeatureStore.from_response_json(json)

        # Act
        fg_res = fs.create_feature_group(
            "test_feature_group_name",
            version=1,
            primary_key=["bob"],
            event_time="when",
            online_enabled=True,
        )

        # Assert
        assert fg_res.name == "test_feature_group_name"
        assert fg_res.version == 1
        assert fg_res.primary_key == ["bob"]
        assert fg_res.event_time == "when"
        assert fg_res.online_enabled is True
        assert fg_res.feature_store == fs
        assert fg_res._feature_store == fs

    def test_get_feature_view_by_name_not_found(self, backend_fixtures, mocker):
        # Arrange
        mocker.patch("hopsworks_common.client.get_instance")
        json = backend_fixtures["feature_store"]["get"]["response"]
        fs = feature_store_mod.FeatureStore.from_response_json(json)
        mocker.patch(
            "hsfs.core.feature_view_api.FeatureViewApi.get_by_name_version", return_value=None
        )

        # Act
        fv_res = fs.get_feature_view("test_feature_view_name")

        # Assert
        assert fv_res is None

    def test_get_feature_views_not_found(self, backend_fixtures, mocker):
        # Arrange
        mocker.patch("hopsworks_common.client.get_instance")
        json = backend_fixtures["feature_store"]["get"]["response"]
        fs = feature_store_mod.FeatureStore.from_response_json(json)
        mocker.patch(
            "hsfs.core.feature_view_api.FeatureViewApi.get_by_name", return_value=[]
        )

        # Act
        fv_res = fs.get_feature_views("test_feature_view_name")

        # Assert
        assert fv_res == []

    def test_get_feature_view_by_name_and_version_not_found(
        self, backend_fixtures, mocker
    ):
        # Arrange
        mocker.patch("hopsworks_common.client.get_instance")
        json = backend_fixtures["feature_store"]["get"]["response"]
        fs = feature_store_mod.FeatureStore.from_response_json(json)
        mocker.patch(
            "hsfs.core.feature_view_api.FeatureViewApi.get_by_name_version",
            return_value=None,
        )

        # Act
        fv_res = fs.get_feature_view("test_feature_view_name")

        # Assert
        assert fv_res is None



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json
from typing import Optional

import humps
from hopsworks_common import util
from hopsworks_common.constants import INFERENCE_BATCHER


class InferenceBatcher:
    """Configuration of an inference batcher for a predictor.

    # Arguments
        enabled: Whether the inference batcher is enabled or not. The default value is `false`.
        max_batch_size: Maximum requests batch size.
        max_latency: Maximum latency for request batching.
        timeout: Maximum waiting time for request batching.
    # Returns
        `InferenceLogger`. Configuration of an inference logger.
    """

    def __init__(
        self,
        enabled: Optional[bool] = None,
        max_batch_size: Optional[int] = None,
        max_latency: Optional[int] = None,
        timeout: Optional[int] = None,
        **kwargs,
    ):
        self._enabled = enabled if enabled is not None else INFERENCE_BATCHER.ENABLED
        self._max_batch_size = max_batch_size if max_batch_size is not None else None
        self._max_latency = max_latency if max_latency is not None else None
        self._timeout = timeout if timeout is not None else None

    def describe(self):
        """Print a description of the inference batcher"""
        util.pretty_print(self)

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        return cls.from_json(json_decamelized)

    @classmethod
    def from_json(cls, json_decamelized):
        return InferenceBatcher(**cls.extract_fields_from_json(json_decamelized))

    @classmethod
    def extract_fields_from_json(cls, json_decamelized):
        config = (
            json_decamelized.pop("batching_configuration")
            if "batching_configuration" in json_decamelized
            else json_decamelized
        )
        kwargs = {}
        kwargs["enabled"] = util.extract_field_from_json(
            config, ["batching_enabled", "enabled"]
        )
        kwargs["max_batch_size"] = util.extract_field_from_json(
            config, "max_batch_size"
        )
        kwargs["max_latency"] = util.extract_field_from_json(config, "max_latency")
        kwargs["timeout"] = util.extract_field_from_json(config, "timeout")

        return kwargs

    def update_from_response_json(self, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        self.__init__(**self.extract_fields_from_json(json_decamelized))
        return self

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        json = {"batchingEnabled": self._enabled}
        if self._max_batch_size is not None:
            json["maxBatchSize"] = self._max_batch_size
        if self._max_latency is not None:
            json["maxLatency"] = self._max_latency
        if self._timeout is not None:
            json["timeout"] = self._timeout
        return {"batchingConfiguration": json}

    @property
    def enabled(self):
        """Whether the inference batcher is enabled or not."""
        return self._enabled

    @enabled.setter
    def enabled(self, enabled: bool):
        self._enabled = enabled

    @property
    def max_batch_size(self):
        """Maximum requests batch size."""
        return self._max_batch_size

    @max_batch_size.setter
    def max_batch_size(self, max_batch_size: int):
        self._max_batch_size = max_batch_size

    @property
    def max_latency(self):
        """Maximum latency."""
        return self._max_latency

    @max_latency.setter
    def max_latency(self, max_latency: int):
        self._max_latency = max_latency

    @property
    def timeout(self):
        """Maximum timeout."""
        return self._timeout

    @timeout.setter
    def timeout(self, timeout: int):
        self._timeout = timeout

    def __repr__(self):
        return f"InferenceBatcher(enabled: {self._enabled!r})"



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


class Default:
    """An object of this class is used in place of optional arguments in
    cases when passing None should result in an exception or other behaviour.
    Overwritting None with a default value is undesirable in some cases,
    and this class solves the problem."""

    def __repr__(self):
        return "DEFAULT"

    def __eq__(self, other):
        return isinstance(other, Default)


DEFAULT = Default()


class JOBS:
    SUCCESS_STATES = ["FINISHED", "SUCCEEDED"]
    ERROR_STATES = [
        "FAILED",
        "KILLED",
        "FRAMEWORK_FAILURE",
        "APP_MASTER_START_FAILED",
        "INITIALIZATION_FAILED",
    ]


class GIT:
    SUCCESS_STATES = ["SUCCESS"]
    ERROR_STATES = ["FAILED", "KILLED", "INITIALIZATION_FAILED", "TIMEDOUT"]


class SERVICES:
    LIST = ["JOBS", "KAFKA", "JUPYTER", "HIVE", "SERVING", "FEATURESTORE", "AIRFLOW"]


class OPENSEARCH_CONFIG:
    SSL_CONFIG = "es.net.ssl"
    NODES_WAN_ONLY = "es.nodes.wan.only"
    NODES = "es.nodes"
    SSL_KEYSTORE_LOCATION = "es.net.ssl.keystore.location"
    SSL_KEYSTORE_PASSWORD = "es.net.ssl.keystore.pass"
    SSL_TRUSTSTORE_LOCATION = "es.net.ssl.truststore.location"
    SSL_TRUSTSTORE_PASSWORD = "es.net.ssl.truststore.pass"
    HTTP_AUTHORIZATION = "es.net.http.header.Authorization"
    INDEX = "es.resource"
    HOSTS = "hosts"
    HTTP_COMPRESS = "http_compress"
    HEADERS = "headers"
    USE_SSL = "use_ssl"
    VERIFY_CERTS = "verify_certs"
    SSL_ASSERT_HOSTNAME = "ssl_assert_hostname"
    CA_CERTS = "ca_certs"


class FEATURES:
    """
    Class that stores constants about a feature.
    """

    MAX_LENGTH_NAME = 63


class KAFKA_SSL_CONFIG:
    """
    Kafka SSL constant strings for configuration
    """

    SSL = "SSL"
    SSL_TRUSTSTORE_LOCATION_CONFIG = "ssl.truststore.location"
    SSL_TRUSTSTORE_LOCATION_DOC = "The location of the trust store file. "
    SSL_TRUSTSTORE_PASSWORD_CONFIG = "ssl.truststore.password"
    SSL_TRUSTSTORE_PASSWORD_DOC = "The password for the trust store file. If a password is not set access to the truststore is still available, but integrity checking is disabled."
    SSL_KEYSTORE_LOCATION_CONFIG = "ssl.keystore.location"
    SSL_KEYSTORE_PASSWORD_CONFIG = "ssl.keystore.password"
    SSL_KEY_PASSWORD_CONFIG = "ssl.key.password"
    SECURITY_PROTOCOL_CONFIG = "security.protocol"
    SSL_CERTIFICATE_LOCATION_CONFIG = "ssl.certificate.location"
    SSL_CA_LOCATION_CONFIG = "ssl.ca.location"
    SSL_PRIVATE_KEY_LOCATION_CONFIG = "ssl.key.location"
    SSL_ENDPOINT_IDENTIFICATION_ALGORITHM_CONFIG = (
        "ssl.endpoint.identification.algorithm"
    )


class KAFKA_PRODUCER_CONFIG:
    """
    Constant strings for Kafka producers
    """

    BOOTSTRAP_SERVERS_CONFIG = "bootstrap.servers"
    KEY_SERIALIZER_CLASS_CONFIG = "key.serializer"
    VALUE_SERIALIZER_CLASS_CONFIG = "value.serializer"


class KAFKA_CONSUMER_CONFIG:
    """
    Constant strings for Kafka consumers
    """

    GROUP_ID_CONFIG = "group.id"
    CLIENT_ID_CONFIG = "client.id"
    ENABLE_AUTO_COMMIT_CONFIG = "enable.auto.commit"
    AUTO_COMMIT_INTERVAL_MS_CONFIG = "auto.commit.interval.ms"
    SESSION_TIMEOUT_MS_CONFIG = "session.timeout.ms"
    KEY_DESERIALIZER_CLASS_CONFIG = "key.deserializer"
    VALUE_DESERIALIZER_CLASS_CONFIG = "value.deserializer"
    AUTO_OFFSET_RESET_CONFIG = "auto.offset.reset"
    ENABLE_AUTO_COMMIT_CONFIG = "enable.auto.commit"
    KEY_DESERIALIZER_CLASS_CONFIG = "key.deserializer"
    VALUE_DESERIALIZER_CLASS_CONFIG = "value.deserializer"


class ENV_VARS:
    """
    Constant strings for environment variables
    """

    KAFKA_BROKERS = "KAFKA_BROKERS"
    ELASTIC_ENDPOINT_ENV_VAR = "ELASTIC_ENDPOINT"


class SSL_CONFIG:
    """
    General SSL configuration constants for Hops-TLS
    """

    KEYSTORE_SUFFIX = "__kstore.jks"
    TRUSTSTORE_SUFFIX = "__tstore.jks"
    PASSWORD_SUFFIX = "__cert.key"

    K_CERTIFICATE_CONFIG = "k_certificate"
    T_CERTIFICATE_CONFIG = "t_certificate"
    PEM_CLIENT_CERTIFICATE_CONFIG = "client.pem"
    PEM_CLIENT_KEY_CONFIG = "client_key.pem"
    PEM_CA_CHAIN_CERTIFICATE_CONFIG = "ca_chain.pem"
    DOMAIN_CA_TRUSTSTORE = "domain_ca_truststore"
    CRYPTO_MATERIAL_PASSWORD = "material_passwd"
    PEM_CA_ROOT_CERT = "/srv/hops/kagent/host-certs/hops_root_ca.pem"
    SSL_ENABLED = "ipc.server.ssl.enabled"


class HOSTS:
    APP_HOST = "c.app.hopsworks.ai"


class MODEL:
    FRAMEWORK_TENSORFLOW = "TENSORFLOW"
    FRAMEWORK_TORCH = "TORCH"
    FRAMEWORK_PYTHON = "PYTHON"
    FRAMEWORK_SKLEARN = "SKLEARN"
    FRAMEWORK_LLM = "LLM"


class MODEL_REGISTRY:
    HOPSFS_MOUNT_PREFIX = "/hopsfs/"
    MODEL_FILES_DIR_NAME = "Files"


class MODEL_SERVING:
    MODELS_DATASET = "Models"
    ARTIFACTS_DIR_NAME = "Artifacts"


class ARTIFACT_VERSION:
    CREATE = "CREATE"


class RESOURCES:
    MIN_NUM_INSTANCES = 1  # disable scale-to-zero by default
    # default values, not hard limits
    MIN_CORES = 0.2
    MIN_MEMORY = 32
    GPUS = 0
    MAX_CORES = 2
    MAX_MEMORY = 1024


class KAFKA_TOPIC:
    NONE = "NONE"
    CREATE = "CREATE"
    NUM_REPLICAS = 1
    NUM_PARTITIONS = 1


class INFERENCE_LOGGER:
    MODE_NONE = "NONE"
    MODE_ALL = "ALL"
    MODE_MODEL_INPUTS = "MODEL_INPUTS"
    MODE_PREDICTIONS = "PREDICTIONS"


class INFERENCE_BATCHER:
    ENABLED = False


class DEPLOYMENT:
    ACTION_START = "START"
    ACTION_STOP = "STOP"


class PREDICTOR:
    # model server
    MODEL_SERVER_PYTHON = "PYTHON"
    MODEL_SERVER_TF_SERVING = "TENSORFLOW_SERVING"
    MODEL_SERVER_VLLM = "VLLM"
    # serving tool
    SERVING_TOOL_DEFAULT = "DEFAULT"
    SERVING_TOOL_KSERVE = "KSERVE"


class PREDICTOR_STATE:
    # status
    STATUS_CREATING = "Creating"
    STATUS_CREATED = "Created"
    STATUS_STARTING = "Starting"
    STATUS_FAILED = "Failed"
    STATUS_RUNNING = "Running"
    STATUS_IDLE = "Idle"
    STATUS_UPDATING = "Updating"
    STATUS_STOPPING = "Stopping"
    STATUS_STOPPED = "Stopped"
    # condition type
    CONDITION_TYPE_STOPPED = "STOPPED"
    CONDITION_TYPE_SCHEDULED = "SCHEDULED"
    CONDITION_TYPE_INITIALIZED = "INITIALIZED"
    CONDITION_TYPE_STARTED = "STARTED"
    CONDITION_TYPE_READY = "READY"


class INFERENCE_ENDPOINTS:
    # endpoint type
    ENDPOINT_TYPE_KUBE_CLUSTER = "KUBE_CLUSTER"
    ENDPOINT_TYPE_LOAD_BALANCER = "LOAD_BALANCER"
    # port name
    PORT_NAME_HTTP = "HTTP"
    PORT_NAME_HTTPS = "HTTPS"
    PORT_NAME_STATUS_PORT = "STATUS"
    PORT_NAME_TLS = "TLS"
    # protocol
    API_PROTOCOL_REST = "REST"
    API_PROTOCOL_GRPC = "GRPC"


class DEPLOYABLE_COMPONENT:
    PREDICTOR = "predictor"
    TRANSFORMER = "transformer"



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from hopsworks_common import client, git_op_execution


class GitOpExecutionApi:
    def _get_execution(self, repo_id, execution_id):
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "git",
            "repository",
            str(repo_id),
            "execution",
            str(execution_id),
        ]
        query_params = {"expand": "repository"}

        return git_op_execution.GitOpExecution.from_response_json(
            _client._send_request("GET", path_params, query_params=query_params)
        )



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import asyncio
import re
from typing import Any, Dict, Optional

from hopsworks_common.core import variable_api
from hsfs.core.constants import HAS_AIOMYSQL, HAS_SQLALCHEMY


if HAS_SQLALCHEMY:
    from sqlalchemy import create_engine
    from sqlalchemy.engine.url import make_url

if HAS_AIOMYSQL:
    from aiomysql.sa import create_engine as async_create_engine


def create_mysql_engine(
    online_conn: Any, external: bool, options: Optional[Dict[str, Any]] = None
) -> Any:
    online_options = online_conn.spark_options()
    # Here we are replacing the first part of the string returned by Hopsworks,
    # jdbc:mysql:// with the sqlalchemy one + username and password
    # useSSL and allowPublicKeyRetrieval are not valid properties for the pymysql driver
    # to use SSL we'll have to something like this:
    # ssl_args = {'ssl_ca': ca_path}
    # engine = create_engine("mysql+pymysql://<user>:<pass>@<addr>/<schema>", connect_args=ssl_args)
    if external:
        # This only works with external clients.
        # Hopsworks clients should use the storage connector
        host = variable_api.VariableApi().get_loadbalancer_external_domain("mysqld")
        online_options["url"] = re.sub(
            "://[0-9a-zA-Z.]+:",
            "://{}:".format(host),
            online_options["url"],
        )

    sql_alchemy_conn_str = (
        online_options["url"]
        .replace(
            "jdbc:mysql://",
            "mysql+pymysql://"
            + online_options["user"]
            + ":"
            + online_options["password"]
            + "@",
        )
        .replace("useSSL=false&", "")
        .replace("?allowPublicKeyRetrieval=true", "")
    )
    if options is not None and not isinstance(options, dict):
        raise TypeError("`options` should be a `dict` type.")
    if not options:
        options = {"pool_recycle": 3600}
    elif "pool_recycle" not in options:
        options["pool_recycle"] = 3600
    # default connection pool size kept by engine is 5
    sql_alchemy_engine = create_engine(sql_alchemy_conn_str, **options)
    return sql_alchemy_engine


async def create_async_engine(
    online_conn: Any,
    external: bool,
    default_min_size: int,
    options: Optional[Dict[str, Any]] = None,
    hostname: Optional[str] = None,
) -> Any:
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError as er:
        raise RuntimeError(
            "Event loop is not running. Please invoke this co-routine from a running loop or provide an event loop."
        ) from er

    online_options = online_conn.spark_options()
    # read the keys user, password from online_conn as use them while creating the connection pool
    url = make_url(online_options["url"].replace("jdbc:", ""))
    if hostname is None:
        if external:
            hostname = variable_api.VariableApi().get_loadbalancer_external_domain(
                "mysqld"
            )
        else:
            hostname = url.host

    if options is None:
        options = {}

    # create a aiomysql connection pool
    pool = await async_create_engine(
        host=hostname,
        port=3306,
        user=online_options["user"],
        password=online_options["password"],
        db=url.database,
        loop=loop,
        minsize=options.get("minsize", default_min_size),
        maxsize=options.get("maxsize", default_min_size),
        pool_recycle=options.get("pool_recycle", 14400),
        autocommit=options.get("autocommit", True),
    )
    return pool



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import Dict, List, Union

from hopsworks_common import client
from hsfs import ge_validation_result


class ValidationResultApi:
    def __init__(self, feature_store_id: int, feature_group_id: int) -> None:
        """Validation Result endpoints for the featuregroup resource.

        :param feature_store_id: id of the respective featurestore
        :type feature_store_id: int
        :param feature_group_id: id of the respective featuregroup
        :type feature_group_id: int
        """
        self._feature_store_id = feature_store_id
        self._feature_group_id = feature_group_id

    def get_validation_history(
        self,
        expectation_id: int,
        query_params: Dict[str, str] = None,
    ) -> Union[
        List[ge_validation_result.ValidationResult],
        ge_validation_result.ValidationResult,
    ]:
        """Get the validation report attached to a featuregroup.

        :return: validation report
        :rtype: Union[List[ValidationResult], ValidationResult]
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            self._feature_store_id,
            "featuregroups",
            self._feature_group_id,
            "validationresult",
            "history",
            expectation_id,
        ]
        headers = {"content-type": "application/json"}

        return ge_validation_result.ValidationResult.from_response_json(
            _client._send_request("GET", path_params, query_params, headers=headers)
        )



#
#   Copyright 2021 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import warnings
from typing import List, Optional

import humps
from hopsworks_common import usage, util
from hsml import model
from hsml.core import model_api
from hsml.llm import signature as llm_signature  # noqa: F401
from hsml.python import signature as python_signature  # noqa: F401
from hsml.sklearn import signature as sklearn_signature  # noqa: F401
from hsml.tensorflow import signature as tensorflow_signature  # noqa: F401
from hsml.torch import signature as torch_signature  # noqa: F401


class ModelRegistry:
    DEFAULT_VERSION = 1

    def __init__(
        self,
        project_name,
        project_id,
        model_registry_id,
        shared_registry_project_name=None,
        **kwargs,
    ):
        self._project_name = project_name
        self._project_id = project_id

        self._shared_registry_project_name = shared_registry_project_name
        self._model_registry_id = model_registry_id

        self._model_api = model_api.ModelApi()

        self._tensorflow = tensorflow_signature
        self._python = python_signature
        self._sklearn = sklearn_signature
        self._torch = torch_signature
        self._llm = llm_signature

        tensorflow_signature._mr = self
        python_signature._mr = self
        sklearn_signature._mr = self
        torch_signature._mr = self
        llm_signature._mr = self

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        return cls(**json_decamelized)

    @usage.method_logger
    def get_model(self, name: str, version: int = None) -> Optional[model.Model]:
        """Get a model entity from the model registry.
        Getting a model from the Model Registry means getting its metadata handle
        so you can subsequently download the model directory.

        # Arguments
            name: Name of the model to get.
            version: Version of the model to retrieve, defaults to `None` and will
                return the `version=1`.
        # Returns
            `Model`: The model metadata object or `None` if it does not exist.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If unable to retrieve model from the model registry.
        """

        if version is None:
            warnings.warn(
                "No version provided for getting model `{}`, defaulting to `{}`.".format(
                    name, self.DEFAULT_VERSION
                ),
                util.VersionWarning,
                stacklevel=1,
            )
            version = self.DEFAULT_VERSION

        return self._model_api.get(
            name,
            version,
            self.model_registry_id,
            shared_registry_project_name=self.shared_registry_project_name,
        )

    @usage.method_logger
    def get_models(self, name: str) -> List[model.Model]:
        """Get all model entities from the model registry for a specified name.
        Getting all models from the Model Registry for a given name returns a list of model entities, one for each version registered under
        the specified model name.

        # Arguments
            name: Name of the model to get.
        # Returns
            `List[Model]`: A list of model metadata objects.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If unable to retrieve model versions from the model registry.
        """

        return self._model_api.get_models(
            name,
            self.model_registry_id,
            shared_registry_project_name=self.shared_registry_project_name,
        )

    @usage.method_logger
    def get_best_model(
        self, name: str, metric: str, direction: str
    ) -> Optional[model.Model]:
        """Get the best performing model entity from the model registry.
        Getting the best performing model from the Model Registry means specifying in addition to the name, also a metric
        name corresponding to one of the keys in the training_metrics dict of the model and a direction. For example to
        get the model version with the highest accuracy, specify metric='accuracy' and direction='max'.

        # Arguments
            name: Name of the model to get.
            metric: Name of the key in the training metrics field to compare.
            direction: 'max' to get the model entity with the highest value of the set metric, or 'min' for the lowest.
        # Returns
            `Model`: The model metadata object or `None` if it does not exist.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If unable to retrieve model from the model registry.
        """

        model = self._model_api.get_models(
            name,
            self.model_registry_id,
            shared_registry_project_name=self.shared_registry_project_name,
            metric=metric,
            direction=direction,
        )
        if isinstance(model, list) and len(model) > 0:
            return model[0]
        else:
            return None

    @property
    def project_name(self):
        """Name of the project the registry is connected to."""
        return self._project_name

    @property
    def project_path(self):
        """Path of the project the registry is connected to."""
        return "/Projects/{}".format(self._project_name)

    @property
    def project_id(self):
        """Id of the project the registry is connected to."""
        return self._project_id

    @property
    def shared_registry_project_name(self):
        """Name of the project the shared model registry originates from."""
        return self._shared_registry_project_name

    @property
    def model_registry_id(self):
        """Id of the model registry."""
        return self._model_registry_id

    @property
    def tensorflow(self):
        """Module for exporting a TensorFlow model."""

        return tensorflow_signature

    @property
    def sklearn(self):
        """Module for exporting a sklearn model."""

        return sklearn_signature

    @property
    def torch(self):
        """Module for exporting a torch model."""

        return torch_signature

    @property
    def python(self):
        """Module for exporting a generic Python model."""

        return python_signature

    @property
    def llm(self):
        """Module for exporting a Large Language Model."""

        return llm_signature

    def __repr__(self):
        project_name = (
            self._shared_registry_project_name
            if self._shared_registry_project_name is not None
            else self._project_name
        )
        return f"ModelRegistry(project: {project_name!r})"



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from hopsworks_common.client import (
    _is_external,
    auth,
    base,
    exceptions,
    external,
    get_instance,
    get_knative_domain,
    get_serving_num_instances_limits,
    hopsworks,
    init,
    is_kserve_installed,
    is_saas_connection,
    is_scale_to_zero_required,
    istio,
    online_store_rest_client,
    set_knative_domain,
    set_kserve_installed,
    set_serving_num_instances_limits,
    stop,
)


__all__ = [
    "auth",
    "base",
    "exceptions",
    "external",
    "get_instance",
    "get_knative_domain",
    "get_serving_num_instances_limits",
    "hopsworks",
    "init",
    "is_kserve_installed",
    "is_saas_connection",
    "is_scale_to_zero_required",
    "istio",
    "online_store_rest_client",
    "set_knative_domain",
    "set_kserve_installed",
    "set_serving_num_instances_limits",
    "stop",
    "_is_external",
]



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import getpass
import json
from typing import List, Optional

from hopsworks_common import client, decorators, secret
from hopsworks_common.core import project_api


class SecretsApi:
    def __init__(
        self,
    ):
        self._project_api = project_api.ProjectApi()

    def get_secrets(self) -> List[secret.Secret]:
        """Get all secrets

        # Returns
            `List[Secret]`: List of all accessible secrets
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()
        path_params = [
            "users",
            "secrets",
        ]
        return secret.Secret.from_response_json(
            _client._send_request("GET", path_params)
        )

    @decorators.catch_not_found(
        "hopsworks_common.secret.Secret", fallback_return=None
    )
    def get_secret(self, name: str, owner: str = None) -> Optional[secret.Secret]:
        """Get a secret.

        # Arguments
            name: Name of the secret.
            owner: username of the owner for a secret shared with the current project. Users can find their username in the Account Settings > Profile section.
        # Returns
            `Secret`: The Secret object or `None` if it does not exist.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()
        query_params = None
        if owner is None:
            path_params = [
                "users",
                "secrets",
                name,
            ]
        else:
            query_params = {"name": name, "owner": owner}
            path_params = [
                "users",
                "secrets",
                "shared",
            ]

        return secret.Secret.from_response_json(
            _client._send_request("GET", path_params, query_params=query_params)
        )[0]

    def get(self, name: str, owner: str = None) -> str:
        """Get the secret's value.
        If the secret does not exist, it prompts the user to create the secret if the application is running interactively

        # Arguments
            name: Name of the secret.
            owner: email of the owner for a secret shared with the current project.
        # Returns
            `str`: The secret value
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        secret_obj = self.get_secret(name=name, owner=owner)
        if secret_obj:
            return secret_obj.value
        else:
            secret_input = getpass.getpass(
                prompt="\nCould not find secret, enter value here to create it: "
            )
        return self.create_secret(name, secret_input).value

    def create_secret(
        self, name: str, value: str, project: str = None
    ) -> secret.Secret:
        """Create a new secret.

        ```python

        import hopsworks

        project = hopsworks.login()

        secrets_api = hopsworks.get_secrets_api()

        secret = secrets_api.create_secret("my_secret", "Fk3MoPlQXCQvPo")

        ```
        # Arguments
            name: Name of the secret.
            value: The secret value.
            project: Name of the project to share the secret with.
        # Returns
            `Secret`: The Secret object
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()

        secret_config = {"name": name, "secret": value}

        if project is None:
            secret_config["visibility"] = "PRIVATE"
        else:
            scope_project = self._project_api._get_project(project)
            secret_config["scope"] = scope_project.id
            secret_config["visibility"] = "PROJECT"

        path_params = [
            "users",
            "secrets",
        ]

        headers = {"content-type": "application/json"}

        _client._send_request(
            "POST", path_params, headers=headers, data=json.dumps(secret_config)
        )

        created_secret = self.get_secret(name)
        print(f"Secret created successfully, explore it at {created_secret.get_url()}")
        return created_secret

    def _delete(self, name: str):
        """Delete the secret.
        :param name: name of the secret
        :type name: Secret
        """
        _client = client.get_instance()
        path_params = [
            "users",
            "secrets",
            name,
        ]
        _client._send_request("DELETE", path_params)



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json

import humps
from hopsworks_common import util


class AlertStatus:
    def __init__(self, inhibited_by=None, silenced_by=None, state=None, **kwargs):
        self._inhibitedBy = inhibited_by
        self._silencedBy = silenced_by
        self._state = state

    @classmethod
    def from_response_json(cls, json_dict):
        if json_dict:
            json_decamelized = humps.decamelize(json_dict)
            return cls(**json_decamelized)
        else:
            return None

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        return {
            "inhibitedBy": self._inhibitedBy,
            "silencedBy": self._silencedBy,
            "state": self._state,
        }

    def __str__(self):
        return self.json()

    def __repr__(self):
        return (
            f"AlertStatus({self._state!r}, {self._inhibitedBy!r}, {self._silencedBy!r})"
        )


class TriggeredAlert:
    def __init__(
        self,
        labels=None,
        annotations=None,
        receivers=None,
        status=None,
        fingerprint=None,
        starts_at=None,
        ends_at=None,
        updated_at=None,
        generatorurl=None,
        **kwargs,
    ):
        self._labels = labels
        self._annotations = annotations
        self._receivers = receivers
        self._fingerprint = fingerprint
        self._starts_at = starts_at
        self._ends_at = ends_at
        self._updated_at = updated_at
        self._generatorurl = generatorurl
        self._status = AlertStatus.from_response_json(status)

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        if "count" in json_decamelized:
            if "items" in json_decamelized:
                return [cls(**receiver) for receiver in json_decamelized["items"]]
            else:
                return []
        else:
            return cls(**json_decamelized)

    @property
    def labels(self):
        """return the labels of the alert"""
        return self._labels

    @property
    def annotations(self):
        """return the annotations of the alert"""
        return self._annotations

    @property
    def receivers(self):
        """return the receivers of the alert"""
        return self._receivers

    @property
    def fingerprint(self):
        """return the fingerprint of the alert"""
        return self._fingerprint

    @property
    def starts_at(self):
        """return the start time of the alert"""
        return self._starts_at

    @property
    def ends_at(self):
        """return the end time of the alert"""
        return self._ends_at

    @property
    def updated_at(self):
        """return the update time of the alert"""
        return self._updated_at

    @property
    def generatorurl(self):
        """return the generator URL of the alert"""
        return self._generatorurl

    @property
    def status(self):
        """return the status of the alert"""
        return self._status

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        return {
            "labels": self._labels,
            "annotations": self._annotations,
            "receivers": self._receivers,
            "fingerprint": self._fingerprint,
            "starts_at": self._starts_at,
            "ends_at": self._ends_at,
            "updated_at": self._updated_at,
            "generatorurl": self._generatorurl,
            "status": self._status.json(),
        }

    def __str__(self):
        return self.json()

    def __repr__(self):
        return f"TriggeredAlert({self._labels!r}, {self._annotations!r}, {self._receivers!r}, {self._status!r}, {self._fingerprint!r}, {self._starts_at!r}, {self._ends_at!r}, {self._updated_at!r}, {self._generatorurl!r})"



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import TypeVar, Union

import hopsworks_common.connection
from hsfs.client import exceptions
from hsfs.engine import spark, spark_no_metastore


_engine = None
_engine_type = None


def init(engine_type: str) -> None:
    global _engine, _engine_type
    python_types = ["python", "training"]
    if _engine_type != engine_type:
        if engine_type in python_types and _engine_type in python_types:
            _engine_type = engine_type
        else:
            stop()
    if not _engine:
        if engine_type == "spark":
            _engine = spark.Engine()
        elif engine_type == "hive":
            raise ValueError(
                "Hive engine is not supported in hopsworks client version >= 4.0."
            )
        elif engine_type == "spark-no-metastore" or engine_type == "spark-delta":
            _engine = spark_no_metastore.Engine()
        elif engine_type in python_types:
            try:
                from hsfs.engine import python
            except ImportError as err:
                raise exceptions.FeatureStoreException(
                    "Trying to instantiate Python as engine, but 'python' extras are "
                    "missing in HSFS installation. Install with `pip install "
                    "hsfs[python]`."
                ) from err
            _engine = python.Engine()
        if _engine:
            _engine_type = engine_type


def get_instance() -> (
    Union[spark.Engine, spark_no_metastore.Engine, TypeVar("python.Engine")]
):
    init(hopsworks_common.connection._hsfs_engine_type)
    return _engine


# Used for testing
def set_instance(
    engine_type: str,
    engine: Union[spark.Engine, spark_no_metastore.Engine, TypeVar("python.Engine")],
) -> None:
    global _engine
    hopsworks_common.connection._hsfs_engine_type = engine_type
    _engine = engine


def get_type() -> str:
    if _engine:
        return hopsworks_common.connection._hsfs_engine_type
    raise Exception("Couldn't find execution engine. Try reconnecting to Hopsworks.")


def stop() -> None:
    global _engine
    from hsfs.core import arrow_flight_client

    _engine = None
    arrow_flight_client.close()



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from abc import abstractmethod

from hopsworks_common.client import base
from hopsworks_common.client.istio.grpc.inference_client import (
    GRPCInferenceServerClient,
)


class Client(base.Client):
    SERVING_API_KEY = "SERVING_API_KEY"
    HOPSWORKS_PUBLIC_HOST = "HOPSWORKS_PUBLIC_HOST"
    TOKEN_EXPIRED_MAX_RETRIES = 0

    BASE_PATH_PARAMS = []

    @abstractmethod
    def __init__(self):
        """To be implemented by clients."""
        pass

    def _get_verify(self, verify, trust_store_path):
        """Get verification method for sending inference requests to Istio.

        Credit to https://gist.github.com/gdamjan/55a8b9eec6cf7b771f92021d93b87b2c

        :param verify: perform hostname verification, 'true' or 'false'
        :type verify: str
        :param trust_store_path: path of the truststore locally if it was uploaded manually to
            the external environment such as EKS or AKS
        :type trust_store_path: str
        :return: if verify is true and the truststore is provided, then return the trust store location
                 if verify is true but the truststore wasn't provided, then return true
                 if verify is false, then return false
        :rtype: str or boolean
        """
        if verify == "true":
            if trust_store_path is not None:
                return trust_store_path
            else:
                return True

        return False

    def _get_host_port_pair(self):
        """
        Removes "http or https" from the rest endpoint and returns a list
        [endpoint, port], where endpoint is on the format /path.. without http://

        :return: a list [endpoint, port]
        :rtype: list
        """
        endpoint = self._base_url
        if endpoint.startswith("http"):
            last_index = endpoint.rfind("/")
            endpoint = endpoint[last_index + 1 :]
        host, port = endpoint.split(":")
        return host, port

    def _close(self):
        """Closes a client. Can be implemented for clean up purposes, not mandatory."""
        self._connected = False

    def _create_grpc_channel(self, service_hostname: str) -> GRPCInferenceServerClient:
        return GRPCInferenceServerClient(
            url=self._host + ":" + str(self._port),
            channel_args=(("grpc.ssl_target_name_override", service_hostname),),
            serving_api_key=self._auth._token,
        )



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json

import humps
from hsfs import util


class TrainingDatasetSplit:
    TIME_SERIES_SPLIT = "TIME_SERIES_SPLIT"
    RANDOM_SPLIT = "RANDOM_SPLIT"
    TRAIN = "train"
    VALIDATION = "validation"
    TEST = "test"

    def __init__(
        self,
        name,
        split_type,
        percentage=None,
        start_time=None,
        end_time=None,
        **kwargs,
    ):
        self._name = name
        self._percentage = percentage
        self._split_type = split_type
        self._start_time = start_time
        self._end_time = end_time

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def percentage(self):
        return self._percentage

    @percentage.setter
    def percentage(self, percentage):
        self._percentage = percentage

    @property
    def split_type(self):
        return self._split_type

    @split_type.setter
    def split_type(self, split_type):
        self._split_type = split_type

    @property
    def start_time(self):
        return self._start_time

    @start_time.setter
    def start_time(self, start_time):
        self._start_time = start_time

    @property
    def end_time(self):
        return self._end_time

    @end_time.setter
    def end_time(self, end_time):
        self._end_time = end_time

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        return {
            "name": self._name,
            "percentage": self._percentage,
            "splitType": self._split_type,
            "startTime": self._start_time,
            "endTime": self._end_time,
        }

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        return cls(
            name=json_decamelized["name"],
            split_type=json_decamelized.get(
                "split_type", TrainingDatasetSplit.RANDOM_SPLIT
            ),
            percentage=json_decamelized.get("percentage", None),
            start_time=json_decamelized.get("start_time", None),
            end_time=json_decamelized.get("end_time", None),
        )



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import logging
import os
import time
import uuid
from datetime import datetime
from typing import Optional

from hopsworks_common.client.exceptions import JobExecutionException, RestAPIError
from hopsworks_common.core import dataset_api, execution_api


class ExecutionEngine:
    def __init__(self):
        self._dataset_api = dataset_api.DatasetApi()
        self._execution_api = execution_api.ExecutionApi()
        self._log = logging.getLogger(__name__)

    def download_logs(self, execution, path=None):
        """Download execution logs to current directory
        :param execution: execution to download logs for
        :type execution: Execution
        :param path: path to download the logs
        :type path: str
        :return: downloaded stdout and stderr log path
        :rtype: str, str
        :raises: JobExecutionException if path is provided but does not exist
        """
        if path is not None and not os.path.exists(path):
            raise JobExecutionException("Path {} does not exist".format(path))
        elif path is None:
            path = os.getcwd()

        job_logs_dir = "logs-job-{}-exec-{}_{}".format(
            execution.job_name, str(execution.id), str(uuid.uuid4())[:16]
        )
        download_log_dir = os.path.join(path, job_logs_dir)

        if not os.path.exists(download_log_dir):
            os.mkdir(download_log_dir)

        out_path = None
        if execution.stdout_path is not None and self._dataset_api.exists(
            execution.stdout_path
        ):
            out_path = self._download_log(execution.stdout_path, download_log_dir)

        err_path = None
        if execution.stderr_path is not None and self._dataset_api.exists(
            execution.stderr_path
        ):
            err_path = self._download_log(execution.stderr_path, download_log_dir)

        return out_path, err_path

    def _download_log(self, path, download_log_dir):
        max_num_retries = 12
        retries = 0
        download_path = None
        while retries < max_num_retries:
            try:
                download_path = self._dataset_api.download(
                    path, download_log_dir, overwrite=True
                )
                break
            except RestAPIError as e:
                if (
                    e.response.json().get("errorCode", "") == 110021
                    and e.response.status_code == 400
                    and retries < max_num_retries
                ):
                    retries += 1
                    time.sleep(5)
                else:
                    raise e
        return download_path

    def wait_until_finished(self, job, execution, timeout: Optional[float] = None):
        """Wait until execution terminates.

        # Arguments
            job: job of the execution
            execution: execution to monitor
            timeout: the maximum waiting time in seconds, if `None` the waiting time is unbounded; defaults to `None`. Note: the actual waiting time may be bigger by approximately 3 seconds.
        # Returns
            `Optional[Execution]`: The final execution or `None` if the timeout is exceeded.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        start_time = datetime.now()

        def passed():
            return (datetime.now() - start_time).total_seconds()

        MAX_LAG = 3.0

        is_yarn_job = job.job_type is not None and (
            job.job_type.lower() == "spark"
            or job.job_type.lower() == "pyspark"
            or job.job_type.lower() == "flink"
        )

        updated_execution = self._execution_api._get(job, execution.id)
        execution_state = None
        while updated_execution.success is None:
            if timeout and timeout <= passed() + MAX_LAG:
                self._log.info("The waiting timeout was exceeded.")
                return updated_execution
            time.sleep(MAX_LAG)
            updated_execution = self._execution_api._get(job, execution.id)
            if execution_state != updated_execution.state:
                if is_yarn_job:
                    self._log.info(
                        "Waiting for execution to finish. Current state: {}. Final status: {}".format(
                            updated_execution.state, updated_execution.final_status
                        )
                    )
                else:
                    self._log.info(
                        "Waiting for execution to finish. Current state: {}".format(
                            updated_execution.state
                        )
                    )
            execution_state = updated_execution.state

        # wait for log files to be aggregated, max 6 minutes
        await_time = 6 * 60.0
        log_aggregation_files_exist = self._dataset_api.exists(
            updated_execution.stdout_path
        ) and self._dataset_api.exists(updated_execution.stderr_path)
        log_aggregation_files_exist_already = log_aggregation_files_exist
        self._log.info("Waiting for log aggregation to finish.")
        while not log_aggregation_files_exist and await_time >= 0:
            if timeout and timeout <= passed() + MAX_LAG:
                break
            await_time -= MAX_LAG
            time.sleep(MAX_LAG)
            updated_execution = self._execution_api._get(job, execution.id)

            log_aggregation_files_exist = self._dataset_api.exists(
                updated_execution.stdout_path
            ) and self._dataset_api.exists(updated_execution.stderr_path)

        if not log_aggregation_files_exist_already:
            if not timeout or timeout > passed() + 5:
                time.sleep(5)  # Helps for log aggregation to flush to filesystem

        if is_yarn_job and not updated_execution.success:
            self._log.error(
                "Execution failed with status: {}. See the logs for more information.".format(
                    updated_execution.final_status
                )
            )
        elif not is_yarn_job and not updated_execution.success:
            self._log.error(
                "Execution failed with status: {}. See the logs for more information.".format(
                    updated_execution.state
                )
            )
        else:
            self._log.info("Execution finished successfully.")

        return updated_execution



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import copy

import humps
from hsml import inference_endpoint


class TestInferenceEndpoint:
    # InferenceEndpointPort

    # from response json

    def test_from_response_json_port(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_endpoint"]["get_port"]["response"]
        json_camelized = humps.camelize(json)  # as returned by the backend
        mock_ie_from_json = mocker.patch(
            "hsml.inference_endpoint.InferenceEndpointPort.from_json"
        )

        # Act
        _ = inference_endpoint.InferenceEndpointPort.from_response_json(json_camelized)

        # Assert
        mock_ie_from_json.assert_called_once_with(json)

    # from json

    def test_from_json_port(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_endpoint"]["get_port"]["response"]
        mock_ie_extract_fields = mocker.patch(
            "hsml.inference_endpoint.InferenceEndpointPort.extract_fields_from_json",
            return_value=json,
        )
        mock_ie_init = mocker.patch(
            "hsml.inference_endpoint.InferenceEndpointPort.__init__", return_value=None
        )

        # Act
        _ = inference_endpoint.InferenceEndpointPort.from_json(json)

        # Assert
        mock_ie_extract_fields.assert_called_once_with(json)
        mock_ie_init.assert_called_once_with(**json)

    # constructor

    def test_constructor_port(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_endpoint"]["get_port"]["response"]

        # Act
        ie_port = inference_endpoint.InferenceEndpointPort(
            name=json["name"], number=json["number"]
        )

        # Assert
        assert ie_port.name == json["name"]
        assert ie_port.number == json["number"]

    # extract fields from json

    def test_extract_fields_from_json_port(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_endpoint"]["get_port"]["response"]
        json_copy = copy.deepcopy(json)

        # Act
        kwargs = inference_endpoint.InferenceEndpointPort.extract_fields_from_json(
            json_copy
        )

        # Assert
        assert kwargs["name"] == json["name"]
        assert kwargs["number"] == json["number"]

    # InferenceEndpoint

    # from response json

    def test_from_response_json_empty(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_endpoint"]["get_empty"]["response"]
        json_camelized = humps.camelize(json)  # as returned by the backend
        mock_ie_from_json = mocker.patch(
            "hsml.inference_endpoint.InferenceEndpoint.from_json"
        )

        # Act
        ie = inference_endpoint.InferenceEndpoint.from_response_json(json_camelized)

        # Assert
        assert isinstance(ie, list)
        assert len(ie) == 0
        mock_ie_from_json.assert_not_called()

    def test_from_response_json_singleton(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_endpoint"]["get_singleton"]["response"]
        json_camelized = humps.camelize(json)  # as returned by the backend
        mock_ie_from_json = mocker.patch(
            "hsml.inference_endpoint.InferenceEndpoint.from_json"
        )

        # Act
        ie = inference_endpoint.InferenceEndpoint.from_response_json(json_camelized)

        # Assert
        assert isinstance(ie, list)
        assert len(ie) == 1
        mock_ie_from_json.assert_called_once_with(json["items"][0])

    def test_from_response_json_list(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_endpoint"]["get_list"]["response"]
        json_camelized = humps.camelize(json)  # as returned by the backend
        mock_ie_from_json = mocker.patch(
            "hsml.inference_endpoint.InferenceEndpoint.from_json"
        )

        # Act
        ie = inference_endpoint.InferenceEndpoint.from_response_json(json_camelized)

        # Assert
        assert isinstance(ie, list)
        assert len(ie) == json["count"]
        assert mock_ie_from_json.call_count == json["count"]

    def test_from_response_json_single(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_endpoint"]["get_singleton"]["response"][
            "items"
        ][0]
        json_camelized = humps.camelize(json)  # as returned by the backend
        mock_ie_from_json = mocker.patch(
            "hsml.inference_endpoint.InferenceEndpoint.from_json"
        )

        # Act
        _ = inference_endpoint.InferenceEndpoint.from_response_json(json_camelized)

        # Assert
        mock_ie_from_json.assert_called_once_with(json)

    # from json

    def test_from_json(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_endpoint"]["get_singleton"]["response"][
            "items"
        ][0]
        mock_ie_extract_fields = mocker.patch(
            "hsml.inference_endpoint.InferenceEndpoint.extract_fields_from_json",
            return_value=json,
        )
        mock_ie_init = mocker.patch(
            "hsml.inference_endpoint.InferenceEndpoint.__init__", return_value=None
        )

        # Act
        _ = inference_endpoint.InferenceEndpoint.from_json(json)

        # Assert
        mock_ie_extract_fields.assert_called_once_with(json)
        mock_ie_init.assert_called_once_with(**json)

    # constructor

    def test_constructor(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_endpoint"]["get_singleton"]["response"][
            "items"
        ][0]

        # Act
        ie = inference_endpoint.InferenceEndpoint(
            type=json["type"], hosts=json["hosts"], ports=json["ports"]
        )

        # Assert
        assert isinstance(ie, inference_endpoint.InferenceEndpoint)
        assert ie.type == json["type"]
        assert ie.hosts == json["hosts"]
        assert ie.ports == json["ports"]

    # extract fields from json

    def test_extract_fields_from_json(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_endpoint"]["get_singleton"]["response"][
            "items"
        ][0]
        json_copy = copy.deepcopy(json)
        mock_ie_port_from_json = mocker.patch(
            "hsml.inference_endpoint.InferenceEndpointPort.from_json", return_value=None
        )

        # Act
        kwargs = inference_endpoint.InferenceEndpoint.extract_fields_from_json(
            json_copy
        )

        # Assert
        assert kwargs["type"] == json["type"]
        assert kwargs["hosts"] == json["hosts"]
        mock_ie_port_from_json.assert_called_once_with(json["ports"][0])

    # get any host

    def test_get_any_host(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_endpoint"]["get_singleton"]["response"][
            "items"
        ][0]
        ie = inference_endpoint.InferenceEndpoint(
            type=None, hosts=json["hosts"], ports=None
        )
        mocker_random_choice = mocker.patch("random.choice", return_value=None)

        # Act
        _ = ie.get_any_host()

        # Assert
        mocker_random_choice.assert_called_once_with(ie.hosts)

    def test_get_any_host_none(self, mocker, backend_fixtures):
        # Arrange
        ie = inference_endpoint.InferenceEndpoint(type=None, hosts=None, ports=None)
        mocker_random_choice = mocker.patch("random.choice", return_value=None)

        # Act
        host = ie.get_any_host()

        # Assert
        assert host is None
        mocker_random_choice.assert_not_called()

    # get port

    def test_get_port_existing(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_endpoint"]["get_list"]["response"]["items"][
            1
        ]
        ports = [
            inference_endpoint.InferenceEndpointPort(p["name"], p["number"])
            for p in json["ports"]
        ]
        ie = inference_endpoint.InferenceEndpoint(type=None, hosts=None, ports=ports)

        # Act
        port = ie.get_port(ports[0].name)

        # Assert
        assert port == ports[0]

    def test_get_port_not_found(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_endpoint"]["get_list"]["response"]["items"][
            1
        ]
        ports = [
            inference_endpoint.InferenceEndpointPort(p["name"], p["number"])
            for p in json["ports"]
        ]
        ie = inference_endpoint.InferenceEndpoint(type=None, hosts=None, ports=ports)

        # Act
        port = ie.get_port("not_found")

        # Assert
        assert port is None

    def test_get_port_none(self):
        # Arrange
        ie = inference_endpoint.InferenceEndpoint(type=None, hosts=None, ports=None)

        # Act
        port = ie.get_port("not_found")

        # Assert
        assert port is None



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

"""Automatic management of aliases.

The associated scripts are located in `python/aliases.py`.
"""

from __future__ import annotations

import functools
import inspect
import warnings
from dataclasses import dataclass, field


class InternalAliasError(Exception):
    """Internal hopsworks exception related to aliases.

    Ideally, this exception should never happen, as it means misconfiguration of aliases, for example, if a public alias is requested for a method of a class.
    """


@dataclass
class Alias:
    from_module: str
    import_name: str
    in_modules: dict[str, InModule] = field(default_factory=dict)

    @dataclass
    class InModule:
        in_module: str
        as_alias: str | None = None
        deprecated_by: set[str] = field(default_factory=set)
        available_until: str | None = None

        def get_id(self):
            res = self.in_module
            if self.as_alias:
                res += f" as {self.as_alias}"
            return res

        def update(self, other: Alias.InModule):
            self.deprecated_by |= other.deprecated_by
            if other.available_until:
                if self.available_until:
                    if self.available_until != other.available_until:
                        raise InternalAliasError(
                            "Deprecated alias is declared available until different releases."
                        )
                else:
                    self.available_until = other.available_until

    def __post_init__(self):
        if "." in self.import_name:
            raise InternalAliasError(
                "Impossible to create alias for not importable symbol."
            )

    def add(self, *in_modules: InModule):
        for im in in_modules:
            self.in_modules.setdefault(im.get_id(), im).update(im)

    def get_id(self):
        return f"{self.from_module}.{self.import_name}"

    def update(self, other: Alias):
        for imid, im in other.in_modules.items():
            self.in_modules.setdefault(imid, im).update(im)


class Registry:
    def __new__(cls):
        return Registry

    aliases: dict[str, Alias] = {}
    modules = {}

    @staticmethod
    def get_modules():
        for module, exclude, paths in Registry.modules.values():
            for name in {x for x, _ in inspect.getmembers(module)} - exclude:
                alias = Alias(module.__name__, name)
                alias.add(*(Alias.InModule(p) for p in paths))
                Registry.add(alias)
        res = {}
        for alias in Registry.aliases.values():
            for im in alias.in_modules.values():
                from_import = alias.from_module, alias.import_name, im
                parts = im.in_module.split(".")
                for i in range(1, len(parts)):
                    res.setdefault(".".join(parts[:i]), [])
                res.setdefault(im.in_module, []).append(from_import)
        return res

    @staticmethod
    def add(*aliases):
        for alias in aliases:
            Registry.aliases.setdefault(alias.get_id(), alias).update(alias)

    @staticmethod
    def add_module(module, exclude, paths):
        Registry.modules[module.__name__] = module, exclude, paths


def public(
    *paths: str,
    as_alias: str | None = None,
    deprecated_by: set[str] | str | None = None,
    available_until: str | None = None,
):
    """Make a function or class publically available.

    If you want to publish a constant, use `publish`.

    # Arguments
        paths: the import paths under which the entity is publically available; effectively results in generation of aliases in all of the paths for the entity.
        as_alias: make the alias of the specified name.
        deprecated_by: make the alias deprecated and provide a set of recommendations to use instead; use of the entity outside hopsworks will print a warning, saying that it is going to be removed from the public API in one of the future releases. See `deprecated` decorator for the implementation of construction of the deprecated objects.
        available_until: the first hopsworks release in which the entity will become unavailable, defaults to `None`; if the release is known, it is reported to the external user in the deprecation warning.
    """

    if available_until and not deprecated_by:
        raise InternalAliasError(
            "If the alias is available until a specific release, something to use instead should be provided in deprecated_by."
        )
    if isinstance(deprecated_by, str):
        deprecated_by = {deprecated_by}

    def decorator(symbol):
        if not hasattr(symbol, "__qualname__"):
            raise InternalAliasError("The symbol should be importable to be public.")
        alias = Alias(symbol.__module__, symbol.__qualname__)
        alias.add(
            *(
                Alias.InModule(p, as_alias, deprecated_by or set(), available_until)
                for p in paths
            )
        )
        Registry.add(alias)
        return symbol

    return decorator


def publish(*paths: str):
    """Publish all of the names defined in this module after this call.

    Since `public` decorator works only for classes and functions, this context should be used for public constants.
    It is also useful for bulk publishing.
    Note that it is impossible to create an alias for a variable, i.e., it is impossible to make a change of a variable in one module to propogate to another variable in another module.

    In case you need to publish something from the begining of a module, use `Publisher`.

    If you need to deprecate an alias, use `public` instead.

    # Arguments
        paths: the import paths under which the names declared in this context will be publically available; effectively results in generation of aliases in all of the paths for all the names declared in the context.
    """

    caller = inspect.getmodule(inspect.stack()[1][0])
    exclude = {x for x, _ in inspect.getmembers(caller)}
    Registry.add_module(caller, exclude, paths)


class Publisher:
    """Publish all of the names defined inside this context.

    This class is intended for bulk publishing of entitities which are to be declared in the begining of a module, so that publish is not usable; in other cases, use publish instead.
    Note that it is impossible to create an alias for a variable, i.e., it is impossible to make a change of a variable in one module to propagate to another variable in another module.
    If you need to deprecate an alias, use `public` instead.

    # Arguments
        paths: the import paths under which the names declared in this context will be publically available; effectively results in generation of aliases in all of the paths for all the names declared in the context.
    """

    def __init__(self, *paths: str):
        self.paths = set(paths)
        caller = inspect.getmodule(inspect.stack()[1][0])
        if caller is None:
            raise InternalAliasError("Cannot publish outside of a module.")
        self.caller = caller

    def __enter__(self):
        self.exclude = {x for x, _ in inspect.getmembers(self.caller)}

    def __exit__(self, _exc_type, _exc_value, _traceback):
        for name in {x for x, _ in inspect.getmembers(self.caller)} - self.exclude:
            alias = Alias(self.caller.__name__, name)
            alias.add(*(Alias.InModule(p) for p in self.paths))
            Registry.add(alias)


class DeprecatedCallWarning(Warning):
    pass


def deprecated(
    *deprecated_by: str,
    available_until: str | None = None,
):
    """Create a deprecated version of a function or a class.

    Use of the entity outside hopsworks will print a warning, saying that it is going to be removed from the public API in one of the future releases.
    Therefore, do not use it on classes or functions used internally; it is a utility for creation of deprecated aliases.

    # Arguments
        deprecated_by: a set of recommendations to use instead.
        available_until: the first hopsworks release in which the entity will become unavailable, defaults to `None`; if the release is known, it is reoprted to the external user in the warning.
    """

    v = f"version {available_until}" if available_until else "a future release"
    if len(deprecated_by) == 1:
        recs = deprecated_by[0]
    elif len(deprecated_by) == 2:
        recs = f"{deprecated_by[0]} or {deprecated_by[1]}"
    else:
        recs = f"{', '.join(deprecated_by[:-1])}, or {deprecated_by[-1]}"

    def deprecate(symbol):
        if inspect.isclass(symbol):
            methods = inspect.getmembers(symbol, predicate=inspect.isfunction)
            for name, value in methods:
                setattr(symbol, name, deprecate(value))
            return symbol
        if inspect.isfunction(symbol):

            @functools.wraps(symbol)
            def deprecated_f(*args, **kwargs):
                caller = inspect.getmodule(inspect.stack()[1][0])
                if not caller or not caller.__name__.startswith("hopsworks"):
                    warnings.warn(
                        f"{symbol.__qualname__} is deprecated."
                        f"The function will be removed in {v} of hopsworks."
                        f"Consider using {recs} instead.",
                        DeprecatedCallWarning,
                        stacklevel=2,
                    )
                return symbol(*args, **kwargs)

            return deprecated_f
        raise InternalAliasError(
            "Deprecation of something else than class or function."
        )

    return deprecate



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional, Union


if TYPE_CHECKING:
    import great_expectations

import humps
from hsfs import util
from hsfs.core.constants import HAS_GREAT_EXPECTATIONS
from hsfs.decorators import uses_great_expectations
from hsfs.ge_validation_result import ValidationResult


if HAS_GREAT_EXPECTATIONS:
    import great_expectations


class ValidationReport:
    """Metadata object representing a validation report generated by Great Expectations in the Feature Store."""

    def __init__(
        self,
        success: bool,
        results: List[
            Union[
                ValidationResult,
                Dict[str, Any],
                great_expectations.core.expectation_validation_result.ExpectationValidationResult,
            ]
        ],
        meta: Optional[Union[Dict[str, Any], str]],
        statistics: Optional[Union[Dict[str, Any], str]],
        evaluation_parameters: Optional[Union[Dict[str, Any], str]] = None,
        id: Optional[int] = None,
        full_report_path: Optional[str] = None,
        featurestore_id: Optional[int] = None,
        featuregroup_id: Optional[int] = None,
        validation_time: Optional[str] = None,
        ingestion_result: Literal[
            "ingested", "rejected", "unknown", "experiment", "fg_data"
        ] = "unknown",
        **kwargs,
    ) -> None:
        self._id = id
        self._success = success
        self._full_report_path = full_report_path
        self._validation_time = validation_time
        self._featurestore_id = featurestore_id
        self._featuregroup_id = featuregroup_id
        self._ingestion_result = ingestion_result

        self.results = results
        self.meta = meta
        self.statistics = statistics
        self.evaluation_parameters = evaluation_parameters

    @classmethod
    def from_response_json(
        cls, json_dict: Dict[str, Any]
    ) -> Union[List[ValidationReport], ValidationReport]:
        json_decamelized = humps.decamelize(json_dict)
        if (
            "count" in json_decamelized
        ):  # todo it has count either way (if items or dict)
            if json_decamelized["count"] == 0:
                return []
            return [
                cls(**validation_report)
                for validation_report in json_decamelized["items"]
            ]
        else:
            return cls(**json_decamelized)

    def json(self) -> str:
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self) -> Dict[str, Union[int, str, bool]]:
        return {
            "id": self._id,
            "success": self.success,
            "evaluationParameters": json.dumps(self.evaluation_parameters),
            "statistics": json.dumps(self._statistics),
            "results": self._results,
            "meta": json.dumps(self._meta),
            "ingestionResult": self._ingestion_result.upper(),
        }

    def to_json_dict(self) -> Dict[str, Any]:
        return {
            "id": self._id,
            "success": self.success,
            "evaluationParameters": self.evaluation_parameters,
            "statistics": self._statistics,
            "results": [result.to_json_dict() for result in self._results],
            "meta": self._meta,
        }

    @uses_great_expectations
    def to_ge_type(self) -> great_expectations.core.ExpectationSuiteValidationResult:
        return great_expectations.core.ExpectationSuiteValidationResult(
            success=self.success,
            statistics=self.statistics,
            results=[result.to_ge_type() for result in self.results],
            evaluation_parameters=self.evaluation_parameters,
            meta=self.meta,
        )

    @property
    def id(self) -> Optional[int]:
        """Id of the validation report, set by backend."""
        return self._id

    @id.setter
    def id(self, id: Optional[int]) -> None:
        self._id = id

    @property
    def success(self) -> bool:
        """Overall success of the validation step"""
        return self._success

    @success.setter
    def success(self, success: bool) -> None:
        self._success = success

    @property
    def results(self) -> List[ValidationResult]:
        """List of expectation results obtained after validation."""
        return self._results

    @results.setter
    def results(
        self,
        results: List[
            Union[
                ValidationResult,
                Dict[str, Any],
                great_expectations.core.expectation_validation_result.ExpectationValidationResult,
            ]
        ],
    ) -> None:
        if len(results) == 0:
            self._results = []
        elif isinstance(results[0], ValidationResult):
            self._results = results
        elif isinstance(results[0], dict):
            self._results = [ValidationResult(**result) for result in results]
        elif isinstance(
            results[0],
            great_expectations.core.expectation_validation_result.ExpectationValidationResult,
        ):
            self._results = [
                ValidationResult(**result.to_json_dict()) for result in results
            ]

    @property
    def meta(self) -> Optional[Dict[str, Any]]:
        """Meta field of the validation report to store additional informations."""
        return self._meta

    @meta.setter
    def meta(self, meta: Optional[Union[Dict[str, Any], str]]) -> None:
        if meta is None:
            self._meta = None
        elif isinstance(meta, dict):
            self._meta = meta
        elif isinstance(meta, str):
            self._meta = json.loads(meta)
        else:
            raise ValueError("Meta field must be stringified json or dict.")

    @property
    def statistics(self) -> Optional[Dict[str, Any]]:
        """Statistics field of the validation report which store overall statistics
        about the validation result, e.g number of failing/successful expectations."""
        return self._statistics

    @statistics.setter
    def statistics(self, statistics: Optional[Union[str, Dict[str, Any]]]) -> None:
        if statistics is None:
            self._statistics = None
        elif isinstance(statistics, dict):
            self._statistics = statistics
        elif isinstance(statistics, str):
            self._statistics = json.loads(statistics)
        else:
            raise ValueError("Statistics field must be stringified json or dict")

    @property
    def evaluation_parameters(self) -> Optional[Dict[str, Any]]:
        """Evaluation parameters field of the validation report which store kwargs of the validation."""
        return self._evaluation_parameters

    @evaluation_parameters.setter
    def evaluation_parameters(
        self, evaluation_parameters: Optional[Union[Dict[str, Any], str]]
    ) -> None:
        if evaluation_parameters is None:
            self._evaluation_parameters = None
        elif isinstance(evaluation_parameters, dict):
            self._evaluation_parameters = evaluation_parameters
        elif isinstance(evaluation_parameters, str):
            self._evaluation_parameters = json.loads(evaluation_parameters)
        else:
            raise ValueError(
                "Evaluation parameters field must be stringified json or dict"
            )

    @property
    def ingestion_result(self) -> str:
        """Overall success of the validation run together with the ingestion
        validation policy. Indicating if dataframe was ingested or rejected."""
        return self._ingestion_result

    @ingestion_result.setter
    def ingestion_result(
        self,
        ingestion_result: Optional[
            Literal["ingested", "rejected", "experiment", "unknown", "fg_data"]
        ],
    ) -> None:
        if ingestion_result is None:
            ingestion_result = "UNKNOWN"
        allowed_values = ["ingested", "rejected", "experiment", "unknown", "fg_data"]
        if ingestion_result.lower() in allowed_values:
            self._ingestion_result = ingestion_result
        else:
            raise ValueError(
                f"Invalid Value {ingestion_result} for ingestion_result."
                + f"Allowed values are {', '.join(allowed_values)}."
            )

    def __str__(self) -> str:
        return self.json()

    def __repr__(self) -> str:
        return (
            f"ValidationReport(success: {self._success}, "
            + f"{self._statistics}, {len(self._results)} results"
            + f" , {self._meta}, {self._full_report_path})"
        )



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


from hsfs import statistics_config


class TestStatistics:
    def test_from_response_json(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["statistics_config"]["get"]["response"]

        # Act
        sc = statistics_config.StatisticsConfig.from_response_json(json)

        # Assert
        assert sc.enabled is True
        assert sc.correlations is True
        assert sc.histograms is True
        assert sc.exact_uniqueness is True
        assert sc.columns == []

    def test_from_response_json_basic_info(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["statistics_config"]["get_basic_info"]["response"]

        # Act
        sc = statistics_config.StatisticsConfig.from_response_json(json)

        # Assert
        assert sc.enabled is True
        assert sc.correlations is False
        assert sc.histograms is False
        assert sc.exact_uniqueness is False
        assert sc.columns == []



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from hopsworks_common.constants import (
    ARTIFACT_VERSION,
    DEFAULT,
    DEPLOYABLE_COMPONENT,
    DEPLOYMENT,
    ENV_VARS,
    GIT,
    HOSTS,
    INFERENCE_BATCHER,
    INFERENCE_ENDPOINTS,
    INFERENCE_LOGGER,
    JOBS,
    KAFKA_CONSUMER_CONFIG,
    KAFKA_PRODUCER_CONFIG,
    KAFKA_SSL_CONFIG,
    KAFKA_TOPIC,
    MODEL,
    MODEL_REGISTRY,
    MODEL_SERVING,
    OPENSEARCH_CONFIG,
    PREDICTOR,
    PREDICTOR_STATE,
    RESOURCES,
    SERVICES,
    SSL_CONFIG,
    Default,
)


__all__ = [
    "ARTIFACT_VERSION",
    "DEFAULT",
    "DEPLOYABLE_COMPONENT",
    "DEPLOYMENT",
    "ENV_VARS",
    "GIT",
    "HOSTS",
    "INFERENCE_BATCHER",
    "INFERENCE_ENDPOINTS",
    "INFERENCE_LOGGER",
    "JOBS",
    "KAFKA_CONSUMER_CONFIG",
    "KAFKA_PRODUCER_CONFIG",
    "KAFKA_SSL_CONFIG",
    "KAFKA_TOPIC",
    "MODEL",
    "MODEL_REGISTRY",
    "MODEL_SERVING",
    "OPENSEARCH_CONFIG",
    "PREDICTOR",
    "PREDICTOR_STATE",
    "RESOURCES",
    "SERVICES",
    "SSL_CONFIG",
    "Default",
]



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Literal, Optional, Union


if TYPE_CHECKING:
    import great_expectations
    import pandas as pd


from hsfs import engine, util, validation_report
from hsfs import expectation_suite as es
from hsfs import feature_group as fg_mod
from hsfs.core.constants import HAS_GREAT_EXPECTATIONS


class GreatExpectationEngine:
    def __init__(self, feature_store_id: int):
        """Engine to run validation using Great Expectations.

        :param feature_store_id: id of the respective Feature Store
        :type `int`
        :return: The engine for Great Expectation objects
        :rtype: `GreatExpectationEngine`
        """
        self._feature_store_id = feature_store_id

    def validate(
        self,
        feature_group: Union[fg_mod.FeatureGroup, fg_mod.ExternalFeatureGroup],
        dataframe: pd.DataFrame,
        expectation_suite: Union[
            great_expectations.core.ExpectationSuite, es.ExpectationSuite, None
        ] = None,
        save_report: bool = False,
        validation_options: Optional[Dict[str, Any]] = None,
        ge_type: bool = True,
        ingestion_result: Literal[
            "unknown", "ingested", "rejected", "fg_data", "experiment"
        ] = "unknown",
    ) -> Union[
        great_expectations.core.ExpectationSuiteValidationResult,
        validation_report.ValidationReport,
        None,
    ]:
        if validation_options is None:
            validation_options = {}

        suite = self.fetch_or_convert_expectation_suite(
            feature_group, expectation_suite, validation_options
        )

        if self.should_run_validation(
            expectation_suite=suite, validation_options=validation_options
        ):
            if not HAS_GREAT_EXPECTATIONS:
                raise ModuleNotFoundError(
                    f"Feature Group {feature_group.name}, v{feature_group.version} is configured to run validation with Great Expectations, "
                    "but Great Expectations is not installed. Please install it using `pip install great_expectations`.\n"
                    "Alternatively you can disable Great Expectations validation by setting `run_validation=False`"
                    "in the validation_options, or disable/delete the suite in the Feature Group Edit UI.\n"
                    f"{util.get_feature_group_url(feature_group.feature_store_id, feature_group.id)}."
                )

            report = engine.get_instance().validate_with_great_expectations(
                dataframe=dataframe,
                expectation_suite=suite.to_ge_type(),
                ge_validate_kwargs=validation_options.get("ge_validate_kwargs", {}),
            )
        else:
            # if run_validation is False we skip validation and saving_report
            return

        if report.success:
            print("Validation succeeded.")
        else:
            print("Validation failed.")
            if (
                suite.validation_ingestion_policy == "STRICT"
                and ingestion_result == "INGESTED"
            ):
                ingestion_result = "REJECTED"

        return self.save_or_convert_report(
            feature_group=feature_group,
            report=report,
            save_report=save_report,
            validation_options=validation_options,
            ingestion_result=ingestion_result,
            ge_type=ge_type,
        )

    def fetch_or_convert_expectation_suite(
        self,
        feature_group: Union[fg_mod.FeatureGroup, fg_mod.ExternalFeatureGroup],
        expectation_suite: Union[
            great_expectations.core.ExpectationSuite, es.ExpectationSuite, None
        ] = None,
        validation_options: Optional[Dict[str, Any]] = None,
    ) -> Optional[es.ExpectationSuite]:
        """Convert provided expectation suite or fetch the one attached to the Feature Group from backend."""
        if expectation_suite is not None:
            if isinstance(expectation_suite, es.ExpectationSuite):
                return expectation_suite
            return es.ExpectationSuite.from_ge_type(expectation_suite)
        if isinstance(validation_options, dict) and not validation_options.get(
            "fetch_expectation_suite", True
        ):
            return feature_group.expectation_suite
        return feature_group.get_expectation_suite(ge_type=False)

    def should_run_validation(
        self,
        expectation_suite: Optional[es.ExpectationSuite],
        validation_options: Dict[str, Any],
    ) -> bool:
        # Suite is None if not provided and nothing attached to FG.
        # In that case we skip validation
        if expectation_suite is None:
            return False

        # If "run_validation" is provided it overrides the value of run_validation of the suite
        return validation_options.get(
            "run_validation", expectation_suite.run_validation
        )

    def save_or_convert_report(
        self,
        feature_group,
        report: great_expectations.core.ExpectationSuiteValidationResult,
        save_report: bool,
        ge_type: bool,
        validation_options: Dict[str, Any],
        ingestion_result: Literal[
            "unknown", "ingested", "rejected", "fg_data", "experiment"
        ] = "unknown",
    ) -> Union[
        great_expectations.core.ExpectationSuiteValidationResult,
        validation_report.ValidationReport,
    ]:
        save_report = validation_options.get("save_report", save_report)
        if save_report:
            return feature_group.save_validation_report(
                report, ingestion_result=ingestion_result, ge_type=ge_type
            )

        if ge_type:
            return report
        else:
            return validation_report.ValidationReport(
                **report.to_json_dict(), ingestion_result=ingestion_result
            )
